"""Tests of gse2sse and sse2sel functions."""
import unittest
from numpy.testing import assert_array_almost_equal_nulp, assert_array_max_ulp, assert_allclose
from copy import deepcopy
from pytplot import data_exists, get_data, store_data, cdf_to_tplot, del_data, tplot_restore, replace_metadata
from pyspedas.themis import gse2sse,sse2sel,autoload_support
from pyspedas import cotrans_get_coord, cotrans_set_coord



class LunCotransDataValidation(unittest.TestCase):
    """ Compares cotrans results between Python and IDL """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: projects/themis/state/cotrans/thm_cotrans_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']

        # Testing tolerance
        cls.tol = 1e-10

        # Download tplot files
        remote_server = 'https://spedas.org/'
        #remote_name = 'testfiles/thm_cotrans_validate.cdf'
        remote_name = 'testfiles/thm_cotrans_validate.tplot'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path=CONFIG['local_data_dir'],
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        #cdf_to_tplot(filename)
        tplot_restore(filename)
        # pytplot.tplot_names()
        # Input variables
        #coord_set_coord('tha_state_pos_gse','gse')
        #coord_set_coord('tha_state_vel_gse','gse')
        cls.tha_state_pos_gse = get_data('tha_state_pos_gse')
        cls.tha_state_vel_gse = get_data('tha_state_vel_gse')
        #coord_set_coord('tha_fgs_gse','gse')
        cls.tha_fgs_gse = get_data('tha_fgs_gse')
        # GSE<->SSE results
        #coord_set_coord('tha_state_pos_sse','sse')
        #coord_set_coord('tha_state_vel_sse','sse')
        cls.tha_state_pos_sse = get_data('tha_state_pos_sse')
        cls.tha_state_vel_sse = get_data('tha_state_vel_sse')
        #coord_set_coord('tha_state_pos_sse_rotate_only','sse')
        #coord_set_coord('tha_state_vel_sse_rotate_only','sse')
        cls.tha_state_pos_sse_rotate_only = get_data('tha_state_pos_sse_rotate_only')
        cls.tha_state_vel_sse_rotate_only = get_data('tha_state_vel_sse_rotate_only')
        #coord_set_coord('tha_fgs_sse','sse')
        cls.tha_fgs_sse = get_data('tha_fgs_sse')
        #coord_set_coord('tha_fgs_sel','sel')
        cls.tha_fgs_sel = get_data('tha_fgs_sel')

        #coord_set_coord('tha_state_pos_gse_sse_gse','gse')
        #coord_set_coord('tha_state_vel_gse_sse_gse','gse')
        cls.tha_state_pos_gse_sse_gse = get_data('tha_state_pos_gse_sse_gse')
        cls.tha_state_vel_gse_sse_gse = get_data('tha_state_vel_gse_sse_gse')

        #coord_set_coord('tha_state_pos_gse_sse_gse_rotate_only','gse')
        #coord_set_coord('tha_state_vel_gse_sse_gse_rotate_only','gse')
        cls.tha_state_pos_gse_sse_gse_rotate_only = get_data('tha_state_pos_gse_sse_gse_rotate_only')
        cls.tha_state_vel_gse_sse_gse_rotate_only = get_data('tha_state_vel_gse_sse_gse_rotate_only')

        #coord_set_coord('tha_fgs_gse_sse_gse','gse')
        cls.tha_fgs_gse_sse_gse = get_data('tha_fgs_gse_sse_gse')

        # SSE<->SSL results
        #coord_set_coord('tha_state_pos_sel','sel')
        cls.tha_state_pos_sel = get_data('tha_state_pos_sel')

        #coord_set_coord('tha_state_pos_gse_sel_sse','sse')
        #coord_set_coord('tha_state_vel_gse_sel_sse','sse')
        cls.tha_state_pos_gse_sel_sse = get_data('tha_state_pos_gse_sel_sse')
        cls.sse_mat_cotrans = get_data('sse_mat_cotrans')
        cls.sel_mat_cotrans = get_data('sel_mat_cotrans')
        cls.sel_x_gei = get_data('sel_x_gei')
        cls.sel_x_gse = get_data('sel_x_gse')
        cls.sel_x_sse = get_data('sel_x_sse')
        cls.sel_y_sse = get_data('sel_y_sse')
        cls.sel_z_sse = get_data('sel_z_sse')

        autoload_support(varname='tha_state_pos_gse', slp=True)

    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # del_data('*')

    def test_replace_metadata(self):
        data = get_data('tha_state_pos_gse')
        orig_meta = deepcopy(get_data('tha_state_pos_gse',metadata=True))
        orig_coord = cotrans_get_coord('tha_state_pos_gse')
        self.assertEqual(orig_coord.lower(), 'gse')
        store_data('newvar',data={'x':data[0],'y':data[1]})
        replace_metadata('newvar',orig_meta)
        self.assertEqual(cotrans_get_coord('newvar').lower(),'gse')
        orig_meta['data_att']['coord_sys'] = 'goofy'  # won't affect tha_state_pos_gse, should not affect newvar either
        self.assertEqual(cotrans_get_coord('newvar').lower(),'gse')
        self.assertEqual(cotrans_get_coord('tha_state_pos_gse').lower(),'gse')

    def test_gse2sse_pos(self):
        """ Validate gse2sse position transform """
        result = gse2sse('tha_state_pos_gse', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_pos_sse', variable_type='pos')
        self.assertEqual(result,1)
        py_sse_mat_cotrans = get_data('sse_mat_cotrans')
        assert_allclose(py_sse_mat_cotrans.y, self.sse_mat_cotrans.y, atol=1.0e-06)
        pos_sse = get_data('tha_state_pos_sse')
        pos_meta = get_data('tha_state_pos_sse',metadata=True)
        self.assertEqual(pos_meta['data_att']['units'],'km')
        assert_allclose(pos_sse.y, self.tha_state_pos_sse.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_state_pos_sse').lower(),'sse')

    def test_gse2sse_pos_rotate_only(self):
        """ Validate gse2sse position transform """
        result = gse2sse('tha_state_pos_gse', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_pos_sse_rotate_only', variable_type='pos',rotation_only=True)
        self.assertEqual(result,1)
        pos_sse = get_data('tha_state_pos_sse_rotate_only')
        pos_meta = get_data('tha_state_pos_sse',metadata=True)
        self.assertEqual(pos_meta['data_att']['units'],'km')
        assert_allclose(pos_sse.y, self.tha_state_pos_sse_rotate_only.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_state_pos_sse_rotate_only').lower(),'sse')

    def test_gse2sse_vel(self):
        """ Validate gse2sse velocity transform """
        result = gse2sse('tha_state_vel_gse', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_vel_sse',variable_type='vel')
        self.assertEqual(result,1)
        vel_sse = get_data('tha_state_vel_sse')
        vel_meta = get_data('tha_state_vel_sse',metadata=True)
        self.assertEqual(vel_meta['data_att']['units'],'km/s')
        assert_allclose(vel_sse.y, self.tha_state_vel_sse.y, atol=1.0e-03)
        self.assertEqual(cotrans_get_coord('tha_state_vel_sse').lower(),'sse')

    def test_gse2sse_vel_rotate_only(self):
        """ Validate gse2sse position transform """
        result = gse2sse('tha_state_vel_gse', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_vel_sse_rotate_only', variable_type='vel',rotation_only=True)
        self.assertEqual(result,1)
        vel_sse = get_data('tha_state_vel_sse_rotate_only')
        vel_meta = get_data('tha_state_vel_sse',metadata=True)
        self.assertEqual(vel_meta['data_att']['units'],'km/s')
        assert_allclose(vel_sse.y, self.tha_state_vel_sse_rotate_only.y, atol=1.0e-03)
        self.assertEqual(cotrans_get_coord('tha_state_vel_sse_rotate_only').lower(),'sse')

    def test_gse2sse_field(self):
        """ Validate gse2sse field transform """
        result = gse2sse('tha_fgs_gse', 'slp_sun_pos', 'slp_lun_pos', 'tha_fgs_sse')
        self.assertEqual(result, 1)
        fgs_sse = get_data('tha_fgs_sse')
        fgs_meta = get_data('tha_fgs_sse',metadata=True)
        self.assertEqual(fgs_meta['data_att']['units'],'nT')
        assert_allclose(fgs_sse.y, self.tha_fgs_sse.y, atol=1.0e-02)
        self.assertEqual(cotrans_get_coord('tha_fgs_sse').lower(), 'sse')

    def test_sse2gse_pos(self):
        """ Validate sse2gse position transform """
        store_data('tha_state_pos_sse',data={'x':self.tha_state_pos_sse.times, 'y':self.tha_state_pos_sse.y})
        cotrans_set_coord('tha_state_pos_sse','sse')
        before_meta = get_data('tha_state_pos_sse',metadata=True)
        before_meta['data_att']['units'] = 'km'
        result = gse2sse('tha_state_pos_sse', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_pos_gse_sse_gse',isssetogse=True,
                         variable_type='pos')
        self.assertEqual(result,1)
        pos_gse = get_data('tha_state_pos_gse_sse_gse')
        pos_meta = get_data('tha_state_pos_gse_sse_gse',metadata=True)
        self.assertEqual(pos_meta['data_att']['units'],'km')
        assert_allclose(pos_gse.y, self.tha_state_pos_gse_sse_gse.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_state_pos_gse_sse_gse').lower(),'gse')

    def test_sse2gse_pos_rotate_only(self):
        """ Validate sse2gse position transform """
        store_data('tha_state_pos_sse_rotate_only',
                           data={'x':self.tha_state_pos_sse_rotate_only.times, 'y':self.tha_state_pos_sse_rotate_only.y})
        cotrans_set_coord('tha_state_pos_sse_rotate_only','sse')
        result = gse2sse('tha_state_pos_sse_rotate_only', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_pos_gse_sse_gse_rotation_only',isssetogse=True,
                         variable_type='pos', rotation_only=True)
        self.assertEqual(result,1)
        pos_gse = get_data('tha_state_pos_gse_sse_gse_rotation_only')
        assert_allclose(pos_gse.y, self.tha_state_pos_gse_sse_gse_rotate_only.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_state_pos_gse_sse_gse_rotate_only').lower(),'gse')

    def test_sse2gse_vel(self):
        """ Validate sse2gse velocity transform """
        result = gse2sse('tha_state_vel_sse', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_vel_gse_sse_gse',isssetogse=True,
                         variable_type='vel')
        self.assertEqual(result,1)
        vel_gse = get_data('tha_state_vel_gse_sse_gse')
        assert_allclose(vel_gse.y, self.tha_state_vel_gse_sse_gse.y, atol=1.0e-02)
        self.assertEqual(cotrans_get_coord('tha_state_vel_gse_sse_gse').lower(),'gse')

    def test_sse2gse_vel_rotate_only(self):
        """ Validate sse2gse position transform """
        store_data('tha_state_vel_sse_rotate_only',
                           data={'x':self.tha_state_vel_sse_rotate_only.times, 'y':self.tha_state_vel_sse_rotate_only.y})
        cotrans_set_coord('tha_state_vel_sse_rotate_only','sse')
        result = gse2sse('tha_state_vel_sse_rotate_only', 'slp_sun_pos', 'slp_lun_pos', 'tha_state_vel_gse_sse_gse_rotation_only',isssetogse=True,
                         variable_type='pos', rotation_only=True)
        self.assertEqual(result,1)
        vel_gse = get_data('tha_state_vel_gse_sse_gse_rotation_only')
        assert_allclose(vel_gse.y, self.tha_state_vel_gse_sse_gse_rotate_only.y, atol=1.0e-03)
        self.assertEqual(cotrans_get_coord('tha_state_vel_gse_sse_gse_rotate_only').lower(),'gse')

    def test_sse2gse_field(self):
        """ Validate gse2sse field transform """
        result = gse2sse('tha_fgs_sse', 'slp_sun_pos', 'slp_lun_pos', 'tha_fgs_gse_sse_gse',isssetogse=True)
        self.assertEqual(result, 1)
        fgs_gse = get_data('tha_fgs_gse_sse_gse')
        assert_allclose(fgs_gse.y, self.tha_fgs_gse_sse_gse.y, atol=1.0e-02)
        self.assertEqual(cotrans_get_coord('tha_fgs_gse_sse_gse').lower(), 'gse')

    def test_sse2sel_pos(self):
        """ Validate sse2sel position transform """
        result = sse2sel('tha_state_pos_sse', 'slp_sun_pos', 'slp_lun_pos', 'slp_lun_att_x', 'slp_lun_att_z', 'tha_state_pos_sel')
        self.assertEqual(result,1)
        py_sel_x_gse = get_data('slp_lun_att_x_gse')
        assert_allclose(self.sel_x_gse.y,py_sel_x_gse.y,atol=1.0e-06)
        py_sel_x_sse = get_data('sel_x_sse')
        assert_allclose(self.sel_x_sse.y,py_sel_x_sse.y,atol=1.0e-06)
        py_sel_y_sse = get_data('sel_y_sse')
        assert_allclose(self.sel_y_sse.y,py_sel_y_sse.y,atol=1.0e-06)
        py_sel_z_sse = get_data('sel_z_sse')
        assert_allclose(self.sel_z_sse.y,py_sel_z_sse.y,atol=1.0e-06)
        py_sel_mat_cotrans = get_data('sel_mat_cotrans')
        assert_allclose(py_sel_mat_cotrans.y, self.sel_mat_cotrans.y, atol=1.0e-06)
        pos_sel = get_data('tha_state_pos_sel')
        pos_meta = get_data('tha_state_pos_sel',metadata=True)
        self.assertEqual(pos_meta['data_att']['units'],'km')
        assert_allclose(pos_sel.y, self.tha_state_pos_sel.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_state_pos_sel').lower(),'sel')

    def test_sse2sel_fgs(self):
        """ Validate sse2sel field transform """
        result = sse2sel('tha_fgs_sse', 'slp_sun_pos', 'slp_lun_pos', 'slp_lun_att_x', 'slp_lun_att_z', 'tha_fgs_sel')
        self.assertEqual(result,1)
        fgs_sel = get_data('tha_fgs_sel')
        assert_allclose(fgs_sel.y, self.tha_fgs_sel.y, atol=.005)
        self.assertEqual(cotrans_get_coord('tha_fgs_sel').lower(),'sel')

    def test_sel2sse_pos(self):
        """ Validate sel2sse position transform """
        # Restore original baseline input tplot variable
        store_data('tha_state_pos_sel',data={'x':self.tha_state_pos_sel.times, 'y':self.tha_state_pos_sel.y})
        cotrans_set_coord('tha_state_pos_sel','sel')

        result = sse2sel('tha_state_pos_sel', 'slp_sun_pos', 'slp_lun_pos', 'slp_lun_att_x', 'slp_lun_att_z', 'tha_state_pos_sel_sse', isseltosse=True)
        self.assertEqual(result,1)
        pos_sse = get_data('tha_state_pos_gse_sel_sse')
        assert_allclose(pos_sse.y, self.tha_state_pos_gse_sel_sse.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_state_pos_gse_sel_sse').lower(),'sse')

    def test_sel2sse_field(self):
        """ Validate sel2sse field transform """
        # Restore original baseline input tplot variable
        store_data('tha_fgs_sel',data={'x':self.tha_fgs_sel.times, 'y':self.tha_fgs_sel.y})
        cotrans_set_coord('tha_fgs_sel','sel')
        md_before = get_data('tha_fgs_sel',metadata=True)
        md_before['data_att']['units'] = 'nT'
        result = sse2sel('tha_fgs_sel', 'slp_sun_pos', 'slp_lun_pos', 'slp_lun_att_x', 'slp_lun_att_z', 'tha_fgs_sel_sse', isseltosse=True)
        self.assertEqual(result,1)
        fgs_sse = get_data('tha_fgs_sel_sse')
        fgs_meta = get_data('tha_fgs_sel_sse',metadata=True)
        self.assertEqual(fgs_meta['data_att']['units'],'nT')
        assert_allclose(fgs_sse.y, self.tha_fgs_sse.y, atol=0.1)
        self.assertEqual(cotrans_get_coord('tha_fgs_sel_sse').lower(),'sse')

if __name__ == '__main__':
    unittest.main()



import numpy as np


def xyz_to_polar(data, co_latitude=False):
    """
    Convert cartesian coordinates to polar coordinates.

    Parameters
    ----------
    x : numpy.ndarray
        x-component of the vector
    y : numpy.ndarray
        y-component of the vector
    z : numpy.ndarray
        z-component of the vector

    Returns
    -------
    r : numpy.ndarray
        radial component of the vector
    theta : numpy.ndarray
        polar angle of the vector
    phi : numpy.ndarray
        azimuthal angle of the vector
    """
    x = data[:, 0]
    y = data[:, 1]
    z = data[:, 2]
    out = np.zeros(data.shape)
    out[:, 0] = np.sqrt(x**2 + y**2 + z**2)
    out[:, 1] = np.arccos(z/out[:, 0])*180.0/np.pi
    out[:, 2] = np.arctan2(y, x)*180.0/np.pi
    if not co_latitude:
        out[:, 1] = 90.0 - out[:, 1]
    return out



import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG


def load(trange=['2012-10-01', '2012-10-02'], 
         instrument='pws',
         datatype='epd', 
         level='l2', 
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the Akebono mission; this function is not meant 
    to be called directly; instead, see the wrappers:

        pyspedas.akebono.pws
        pyspedas.akebono.rdm
        pyspedas.akebono.orb

    """
    prefix = ''

    if instrument == 'pws':
        #  only PWS data are available in CDF files
        prefix = 'akb_pws_'
        pathformat = instrument + '/NPW-DS/%Y/ak_h1_pws_%Y%m%d_v??.cdf'
    elif instrument == 'rdm':
        prefix = 'akb_rdm_'
        pathformat = instrument + '/%Y/sf%y%m%d'
    elif instrument == 'orb':
        prefix = 'akb_orb_'
        pathformat = 'orbit/daily/%Y%m/ED%y%m%d.txt'
    else:
        logging.error('Unknown instrument: ' + instrument)
        return

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly or instrument != 'pws':
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

    


"""
@Author: Xin Cao, Xiangning Chu, University of Colorado Boulder
@Created in Sep. 2021.

In this config.py file, the following environment variables need to be defined in advance.
@LOCAL_SECS_DATA_DIR: the local path where the SECS/EICS data will be downloaded and stored.
@PLOTS_SECS_DIR: the local path where the vector and contour plots of SECS/EICS data will be saved.
"""

import os
CONFIG ={'local_data_dir': '',
         'remote_data_dir': 'http://vmo.igpp.ucla.edu/data1/SECS/'}

# override local data directory with environment variables
if os.environ.get('LOCAL_SECS_DATA_DIR'):
    CONFIG['local_data_dir']  = os.environ['LOCAL_SECS_DATA_DIR'] # self-defined, e.g. '/Users/cao/Documents/Data_N_Res/data_local'

if os.environ.get('PLOTS_SECS_DIR'):
    CONFIG['plots_dir']  = os.environ['PLOTS_SECS_DIR'] # self-defined, e.g. '/Users/cao/Documents/Data_N_Res/plots'

if os.environ.get('CONDA_PREFIX'):
    env_name = os.environ['CONDA_PREFIX']
    os.environ['PROJ_LIB'] = env_name + '/share/proj'



import logging
import numpy as np
import scipy
from pytplot import get_data, store_data, options

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean


def rbsp_rbspice_pad_spinavg(probe='a', datatype='TOFxEH', level='l3', species=None, energy=[0, 1000], bin_size=15., scopes=None):
    """
    Calculates spin-averaged PADs for the RBSPICE instrument
    
    Parameters
    ----------
    probe : str
        RBSP spacecraft indicator [Options: 'a' (default), 'b']
    datatype : str
        desired data type [Options: 'TOFxEH' (default), 'TOFxEnonH']
    level : str
        data level ['l1','l2','l3' (default),'l3pap']
    species : str
        desired ion species [Options: 'proton' (default), 'helium', 'oxygen']
    energy : list
        user-defined energy range to include in the calculation in keV [default = [0,1000]]
    bin_size : float
        desired size of the pitch angle bins in degrees [default = 15]
    scopes : list
        string array of telescopes to be included in PAD [0-5, default is all]

    Returns
    --------
    Tplot variables created
    """
    if level != 'l1':
        units_label = '1/(cm^2-sr-s-keV)'
    else:
        units_label = 'counts/s'
    if species is None and datatype == 'TOFxEH':
        species = 'proton'
    elif species is None and datatype == 'TOFxEnonH':
        species = ['helium', 'oxygen']
    elif species is None and datatype == 'TOFxPHHHELT':
        species = ['proton', 'oxygen']
    if energy is None:
        energy = [0, 1000]
    if bin_size is None:
        bin_size = 15.
    if scopes is None:
        scopes = [0, 1, 2, 3, 4, 5]
    
    en_range_string = str(energy[0]) + '-' + str(energy[1]) + 'keV'
    
    prefix = 'rbsp'+probe+'_rbspice_'+level+'_'+datatype+'_'
    spin_nums = get_data(prefix + 'Spin')

    if spin_nums is None:
        logging.error('Spin variable not found: ' + prefix + 'Spin')
        return

    # find where the spins start
    spin_starts = np.unique(spin_nums.y, return_index=True)[1][1:]-1
    if len(scopes) == 6:
        pad_name = [prefix+species+'_omni_'+en_range_string+'_pad']
    else:
        pad_name = [prefix+species+'_T'+str(i)+'_'+en_range_string+'_pad' for i in scopes]

    logging.info('Calculating spin averaged pitch angle distribution..')
    out = []

    for ii in range(len(pad_name)):
        pad_data = get_data(pad_name[ii])
        
        if pad_data is None:
            logging.error('Error, variable containing valid PAD data missing.')
            return

        # the following is for rebinning and interpolating to new_bins
        n_pabins = 180. / bin_size
        new_bins = 180. * np.arange(n_pabins + 1) / n_pabins
        srx = [float(len(pad_data.v)) / (int(n_pabins) + 1) * (x + 0.5) - 0.5 for x in range(int(n_pabins) + 1)]

        spin_sum_flux = np.zeros((len(spin_starts), len(pad_data.y[0, :])))
        rebinned_data = np.zeros((len(spin_starts), len(pad_data.y[0, :])+1))
        spin_times = np.zeros(len(spin_starts))
        
        current_start = 0
        # loop through the spins for this telescope
        for spin_idx in range(len(spin_starts)):
            # loop over energies
            spin_sum_flux[spin_idx,:] = nanmean(pad_data.y[current_start:spin_starts[spin_idx]+1,:], axis=0)
            spin_times[spin_idx] = pad_data.times[current_start]
            # rebin the data before storing it
            # the idea here is, for bin_size = 15 deg, rebin the data from center points to:
            #    new_bins = [0, 15, 30, 45, 60, 75, 90, 105, 120, 135 , 150, 165, 180]
            spin_sum_interp = scipy.interpolate.interp1d(np.arange(len(spin_sum_flux[spin_idx, :])), spin_sum_flux[spin_idx, :], fill_value='extrapolate')
            rebinned_data[spin_idx, :] = spin_sum_interp(srx)

            # we want to take the end values instead of extrapolating
            # again, to match the functionality of congrid in IDL
            rebinned_data[spin_idx, 0] = spin_sum_flux[spin_idx, 0]
            rebinned_data[spin_idx, -1] = spin_sum_flux[spin_idx, -1]

            current_start = spin_starts[spin_idx]+1
        
        newname = pad_name[ii]+'_spin'
        if len(scopes) == 6:
            ytitle = 'rbsp-'+probe+'\nrbspice\n'+species+'\nomni'
        else:
            ytitle = 'rbsp-'+probe+'\nrbspice\n'+species+'\nT'+str(scopes[ii])

        store_data(newname, data={'x': spin_times, 'y': rebinned_data, 'v': new_bins})
        options(newname, 'spec', True)
        options(newname, 'zlog', True)
        options(newname, 'ztitle', units_label)
        options(newname, 'ytitle', ytitle)
        options(newname, 'yrange', [0, 180.0])
        options(newname, 'ysubtitle', en_range_string+'\nspin-avg PAD\n(deg)')
        out.append(newname)

        #tdegap(newname, overwrite=True)
    return out



import logging
from copy import deepcopy
import numpy as np
from pyspedas.utilities.interpol import interpol


def qmult(q1, q2):
    """
    multiply two quaternions or two arrays of quaternions

    Parameters
    ----------
    q1 : array_like
        a 4 element array, or an Nx4 element array, representing quaternion(s)
    q2 : array_like
        a 4 element array, or an Nx4 element array, representing quaternion(s)

    Returns
    -------
    q1*q2, or -1 on failure

    Notes
    -----
    Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                     q[1] = vector x
                     q[2] = vector y
                     q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """
    q1i = np.array(q1)
    q2i = np.array(q2)

    if q1i.ndim != q2i.ndim:
        logging.error('Number of dimensions in quaternion q1 and quaternion q2 do not match')
        return -1

    # check to make sure input has the correct dimensions
    q1i = qvalidate(q1i, 'q1', 'qmult')
    q2i = qvalidate(q2i, 'q2', 'qmult')

    if isinstance(q1i, int):
        return q1i

    if isinstance(q2i, int):
        return q2i

    # make sure elements match
    if q1i.size != q2i.size:
        logging.error('Number of elements in quaternion q1 and quaternion q2 do not match')
        return -1

    # now the actual dirty work
    qtmp0 = q1i[:, 0] * q2i[:, 0] - q1i[:, 1] * q2i[:, 1] - q1i[:, 2] * q2i[:, 2] - q1i[:, 3] * q2i[:, 3]
    qtmp1 = q1i[:, 1] * q2i[:, 0] + q1i[:, 0] * q2i[:, 1] - q1i[:, 3] * q2i[:, 2] + q1i[:, 2] * q2i[:, 3]
    qtmp2 = q1i[:, 2] * q2i[:, 0] + q1i[:, 3] * q2i[:, 1] + q1i[:, 0] * q2i[:, 2] - q1i[:, 1] * q2i[:, 3]
    qtmp3 = q1i[:, 3] * q2i[:, 0] - q1i[:, 2] * q2i[:, 1] + q1i[:, 1] * q2i[:, 2] + q1i[:, 0] * q2i[:, 3]

    qout = np.array([qtmp0, qtmp1, qtmp2, qtmp3]).T

    return qout


def qdecompose(q):
    """
    Purpose: decompose quaternions into axes and angeles
    Inputs: q:  a 4 element quaternion or an Nx4 element array of quaternions
    Returns: a 4 element array with a[0] = angle, and a[1:3] = axis, or
    an Nx4 element array or -1L on failure
    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                 q[1] = vector x
                 q[2] = vector y
                 q[3] = vector z
    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs
    As per the euve implementation, if q[0] is outside of the range of
    acos...[-1,1] the value of the quaternion will be turned into an
    identity quaternion...in other words clipped, this seems suspect,
    a better solution may be to wrap the value back into range using
    modular arithmatic, future modifiers of this routine should consider
    adding this.

    This routine is based on the IDL version by Patrick Cruce
    """
    EPSILON = 1.0e-20  # Where sin(theta) is close enough to theta
    # this is to avoid mutating the input variable
    qi = q
    # check to make sure input has the correct dimensions
    qi = qvalidate(qi, 'q', 'qdecompose')
    if isinstance(qi, int):
        return qi
    qdims = qi.shape
    aout = np.zeros(qdims, dtype=np.float64)

    # the following code will clip into range
    idx = np.argwhere(qi[:, 0] >= 1.0).flatten()
    if len(idx) != 0:
        aout[idx, 0] = 0.0
        aout[idx, 1] = 1.0
        aout[idx, 2:3] = 0.0

    idx = np.argwhere(qi[:, 0] <= -1.0).flatten()
    if len(idx) != 0:
        aout[idx, 0] = 2 * np.pi
        aout[idx, 1] = 1.0
        aout[idx, 2:3] = 0.0

    idx = np.argwhere((qi[:, 0] > -1.0) & (qi[:, 0] < 1.0)).flatten()
    if len(idx) != 0:
        theta2 = np.arccos(qi[idx, 0])
        aout[idx, 0] = 2 * theta2

        idx2 = np.argwhere(theta2 < EPSILON).flatten()
        if len(idx2) != 0:
            aout[idx[idx2], 1] = 1.0
            aout[idx[idx2], 2:3] = 0.0

        idx2 = np.argwhere(theta2 >= EPSILON).flatten()
        if len(idx2) != 0:
            aout[idx[idx2], 1] = qi[idx[idx2], 1] / np.sin(theta2[idx2])
            aout[idx[idx2], 2] = qi[idx[idx2], 2] / np.sin(theta2[idx2])
            aout[idx[idx2], 3] = qi[idx[idx2], 3] / np.sin(theta2[idx2])

    return aout.reshape(qdims)


def qvalidate(q, argname, fxname):
    """
    Purpose: validate inputs for the idl quaternion library

    Inputs: q: a 4 element array, or an Nx4 element array, representing quaternion(s)
            argname: the name of the argument to be used in error messages

    Returns: an Nx4 array or -1, it will turn 4 element quaternion arrays
             into 1x4 element quaternion arrays

    Notes: This function is here because I noticed a lot of the error
    checking code was being repeated, and it was making the functions
    long and hard to read

    Represention has q[0] = scalar component
                     q[1] = vector x
                     q[2] = vector y
                     q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """
    # this is to avoid mutating the input variable
    qi = deepcopy(q)

    if isinstance(qi, int):
        return -1

    # check to make sure input has the correct dimensions
    elif np.size(np.shape(qi)) == 1:
        if np.size(qi) != 4:
            logging.error('Wrong number of elements in quaternion ' + argname + '. Found when validating input for ' + fxname)
            return -1
        qi = np.reshape(qi, (1, 4))
    elif np.size(np.shape(qi)) == 2:
        s = np.shape(qi)
        if s[np.size(s)-1] != 4:
            logging.error('Dimension 2 of quaternion ' +argname+' must have 4 elements. Found when validating input for ' + fxname)
            return -1
    else:
        logging.error('Quaternion '+argname+' has the wrong number of dimensions. Found when validating input for ' + fxname)
        return -1

    return qi


def qconj(q):
    """
    Purpose: calculate the conjugate a quaternion or an array of quaternions

    Inputs: q: a 4 element array, or an Nx4 element array, representing quaternion(s)

    Returns: q*

    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                     q[1] = vector x
                     q[2] = vector y
                     q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """
    # this is to avoid mutating the input variable
    qi = deepcopy(q)

    # check to make sure input has the correct dimensions
    qi = qvalidate(qi, 'q', 'qconj')

    if isinstance(qi, int):
        return qi

    # the actual conjugation
    qtmp0 = qi[:, 0]
    qtmp1 = -qi[:, 1]
    qtmp2 = -qi[:, 2]
    qtmp3 = -qi[:, 3]

    qout = np.array([qtmp0, qtmp1, qtmp2, qtmp3]).T

    if len(q.shape) == 1:
        qout = qout.flatten()

    return qout


def qslerp(q, x1, x2, geometric=False, eq_tolerance=1e-12):
    """
    Uses spherical linear interpolation to interpolate quaternions between elements of q

    Parameters
    ----------
    q : array_like
        An Nx4 element array, representing a list of quaternions with N > 1, all quaternions must be unit quaternions(ie length/norm = 1)
    x1 : array_like
        The input abscissa values of the quaternions,an array of length N, abscissa values must also be monotonic
    x2 : array_like
        The output abscissa values for the quaternions, can have as many elements as wanted but must fall on the interval [x[0],x[N-1]], an M element array, abscissa values must also be monotonic
    geometric : bool, optional
        This keyword allows you to specify that it use the geometric formula for the slerp. The default formula is probably faster and more numerically stable, the geometric option is just available for testing
        Testing of the geometric method indicates that the norm of the interpolated quaternions strays easily from unit length, when it renormalizes results may be destabilized
    eq_tolerance : float, optional
        Set to specify the tolerance used when determining whether two numbers are equal (default: 1e-12). This tolerance will be used in checking equivalence of:
            -quaternion lengths
            -input vs. output abscissae
            -quaternion direction (inner product)

    Returns
    -------
    q_out : array_like
        An Mx4 element array of interpolated quaternions or -1L on failure

    Notes
    -----
    Represention has q[0] = scalar component
                 q[1] = vector x
                 q[2] = vector y
                 q[3] = vector z

    The vector component of the quaternion can also be thought of as an eigenvalue of the rotation the quaterion performs

    The scalar component can be thought of as the amount of rotation that the quaternion performs

    While the code may seem a little esoteric, it is vectorized and provides the most accurate results it can get

    This routine is based on the IDL version by Patrick Cruce
    """
    qi = deepcopy(q)
    x2i = deepcopy(x2)
    x1i = deepcopy(x1)

    # check that quaternions are consistent with generic quaternion invariants
    qi = qvalidate(qi,'qi','qslerp')

    if isinstance(qi, int):
        return qi

    # check that input quaternions are unit length
    qn = qnorm(qi)

    idx = np.argwhere(np.abs(qn - 1.0) > eq_tolerance).flatten()
    if len(idx) > 0:
        logging.error('At least one input quaternion is not unit length')
        return

    if qi.shape[0] != len(x1i):
        logging.error('Number of input abscissa values does not match the number of input quaternions')
        return

    # check that input abscissa values are monotonic
    if len(x1i) > 1:
        idx = np.argwhere((x1i[1:len(x1i)]-x1i[0:len(x1i)-1]) < 0)
        if len(idx) > 0:
            logging.error('input abscissa values not monotonic')
            return

    # check that output abscissa values are strictly monotonic
    if len(x2i) > 1:
        idx = np.argwhere((x2i[1:len(x2i)]-x2i[0:len(x2i)-1]) < 0)
        if len(idx) > 0:
            logging.error('output abscissa values not monotonic')
            return

    # construct the output array
    q_out = np.zeros((len(x2i), 4))

    # if output abscissa values are outside of the range of input abscissa
    # values constant extrapolation is used
    idx = np.argwhere(x2i < x1i[0]).flatten()

    if len(idx) > 0:
        q_out[idx, :] = np.array(idx.size*[qi[0, :]])

    idx = np.argwhere(x2i > x1i[-1]).flatten()

    if len(idx) > 0:
        q_out[idx, :] = np.array(idx.size*[qi[-1, :]])

    out_idx = np.argwhere((x2i >= x1i[0]) & (x2i <= x1i[-1])).flatten()

    if len(out_idx) == 0:
        return q_out.reshape((-1, 4))

    x2i = x2i[out_idx]

    # construct arguments to the slerp function, this includes the source
    # quaternion list, the target quaternions list, and the proportion of
    # interpolation list for each quaternion pair.  They should all have
    # the same number of elements as the output abscissa value list

    t_temp = interpol(np.arange(qi.shape[0], dtype='float64'), x1i, x2i)

    t_list = t_temp % 1.0

    q_idx = np.int64(np.floor(t_temp))

    # if the last abscissa values are identical,the indexing scheme to
    # generate the q_list could generate an overflow, the two conditionals
    # below prevent this
    idx = np.argwhere(np.abs(t_list) <= eq_tolerance).flatten()  # where t_list =~ 0.0
    if len(idx) > 0:
        q_out[out_idx[idx], :] = qi[q_idx[idx], :]

    slerp_idx = np.argwhere(np.abs(t_list) > eq_tolerance).flatten()  # where t_list !=~ 0.0

    # if there is nothing left, then we're done
    if slerp_idx.size == 0:
        return q_out.reshape((-1, 4))

    q_idx = q_idx[slerp_idx]
    out_idx = out_idx[slerp_idx]
    t_list = t_list[slerp_idx]

    q1_list = qi[q_idx, :]

    q2_list = qi[q_idx + 1, :]

    # calculate the dot product which is needed to to flip the
    # appropriate quaternions to guarantee interpolation is done along the
    # shortest path
    dotp = qdotp(q1_list, q2_list)

    if dotp.ndim == 0 and dotp == -1:
        return -1

    # the following code flips quaternions in q2_list to ensure the
    # shortest path is followed
    idx = np.argwhere(dotp < 0.0).flatten()

    if idx.size != 0:
        q2_list[idx, :] = -q2_list[idx, :]

    # interpolation cannot be performed on colinear quaternions
    # it is assumed that colinear quaternions will be returned unchanged
    # since dotp(q1,q2) = cos(angle between q1,q2) if dotp = 1.0 the
    # quaternions are colinear
    idx = np.argwhere(np.abs(dotp - 1.0) <= eq_tolerance).flatten()  # where dotp = 1.0

    # store colinear quaternions into output array
    if idx.size != 0:
        q_out[out_idx[idx], :] = q1_list[idx, :]

    # copy non-colinear quaternions for processing
    idx = np.argwhere(np.abs(dotp - 1.0) > eq_tolerance).flatten()

    if idx.size == 0:
        return q_out.reshape((-1, 4))  # if no non-colinear quaternions are left, we are done

    dotp = dotp[idx]
    t_list = t_list[idx]
    q1_list = q1_list[idx, :]
    q2_list = q2_list[idx, :]
    out_idx = out_idx[idx]

    # now the actual processing begins

    # testing both methods to verify results
    if geometric:
        theta = np.arccos(dotp)

        sin_theta = np.sin(theta)

        theta_t = theta * t_list

        co1 = np.sin(theta - theta_t) / sin_theta
        co2 = np.sin(theta_t) / sin_theta

        q_out[out_idx, 0] = co1 * q1_list[:, 0] + co2 * q2_list[:, 0]
        q_out[out_idx, 1] = co1 * q1_list[:, 1] + co2 * q2_list[:, 1]
        q_out[out_idx, 2] = co1 * q1_list[:, 2] + co2 * q2_list[:, 2]
        q_out[out_idx, 3] = co1 * q1_list[:, 3] + co2 * q2_list[:, 3]
    else:
        # slerp will be performed by calculating:
        # ((q2*(q1^-1))^t)*q1
        # since the quaternions are unit q1^-1 = conjugate(q1)
        # exponentiation can be calculated by transforming to
        # polar form cos(theta*t)+v*sin(theta*t)
        # theta = acos(q[0])
        # NOTE: this potentially more numerically stable implementation needs
        # to be verified by comparison to the geometric slerp
        q1_conj = qconj(q1_list)

        q2_q1_prod = qdecompose(qmult(q2_list, q1_conj))

        if isinstance(q2_q1_prod, int):
            return -1

        # sometimes a dimension disappears.
        if q2_q1_prod.ndim == 1 and q2_q1_prod.size == 4:
            q2_q1_prod = q2_q1_prod.reshape((1, 4))

        theta_scale = q2_q1_prod[:, 0] * t_list

        q_total = qmult(qcompose(q2_q1_prod[:, 1:4], theta_scale), q1_list)

        if isinstance(q_total, int):
            return -1

        q_out[out_idx, :] = q_total

    return qnormalize(q_out)


def qdotp(q1, q2):
    """
    Purpose: calculate the dot product of two quaternions or two arrays of quaternions

    Inputs: q1: a 4 element array, or an Nx4 element array, representing quaternion(s)
            q2: a 4 element array, or an Nx4 element array, representing quaternion(s)

    Returns: q1.q2, or -1 on failure

    Notes:
    Represention has q[0] = scalar component
                     q[1] = vector x
                     q[2] = vector y
                     q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    The scalar component can be thought of as the amount of rotation that
    the quaternion performs

    like any vector the if t = the angle between q1 and q2 in 4-space
    the q1.q2 = ||q1||*||q2||*cos(t) where || denotes the norm(length) of
    the quaternion in 4-space

    This routine is based on the IDL version by Patrick Cruce
    """
    q1i = deepcopy(q1)
    q2i = deepcopy(q2)
    qout = np.nansum(q1i*q2i, axis=1)
    return qout


def qnorm(q):
    """
    Purpose: calculate the norm a quaternion or an array of quaternions

    Inputs: q: a 4 element array, or an Nx4 element array, representing quaternion(s)

    Returns: norm(q): sqrt(a^2+b^2+c^2+d^2) or -1L on fail
                      will be a single element or an N length array

    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                 q[1] = vector x
                 q[2] = vector y
                 q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This implementation of norm does not apply the squareroot sometimes
    applied to a norm.  If required the sqrt can easily be applied by the user

    This routine is based on the IDL version by Patrick Cruce

    """
    qi = deepcopy(q)
    dotp = qdotp(qi, qi)
    return np.sqrt(dotp)


def qnormalize(q):
    """
    Purpose: normalize a quaternion or an array of quaternions

    Inputs: q: a 4 element array, or an Nx4 element array, representing quaternion(s)

    Returns: q/(sqrt(norm(q))) or -1L on fail

    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                  q[1] = vector x
                  q[2] = vector y
                  q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """
    qi = deepcopy(q)
    qn = qnorm(qi)

    qtmp0 = qi[:, 0]/qn
    qtmp1 = qi[:, 1]/qn
    qtmp2 = qi[:, 2]/qn
    qtmp3 = qi[:, 3]/qn

    qout = np.array([qtmp0, qtmp1, qtmp2, qtmp3]).transpose()

    idx = np.argwhere(qout[:, 0] > 1.0).flatten()
    if len(idx) > 0:
        qout[idx, 0] = 1.0
        qout[idx, 1:4] = 0.0

    idx = np.argwhere(qout[:, 0] < -1.0).flatten()
    if len(idx) > 0:
        qout[idx, 0] = -1.0
        qout[idx, 1:4] = 0.0

    return qout


def mtoq(m):
    """
    Function: mtoq

    Purpose: transforms a rotation matrix into a quaternion.  If the
    matrix does not perform a rotation, then its behavior may be ill-
    defined

    WARNING!!!! - this routine does not conform to the wikipedia definition.  see warning for qtom.pro

    Inputs: m: a 3x3 element array or an Nx3x3 element array

    Returns: q

    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                     q[1] = vector x
                     q[2] = vector y
                     q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """
    mi = deepcopy(m)

    dims = np.shape(mi)

    if len(dims) == 2:
        if dims[0] != 3 or dims[1] != 3:
            logging.error('Wrong dimensions in input matrix')
            return -1

        mi = np.reshape(m, (1, 3, 3))

        dims = [1, dims]

    elif len(dims) == 3:
        if dims[1] != 3 or dims[2] != 3:
            logging.error('Wrong dimensions in input matrix')
            return -1
    else:
        logging.error('Wrong dimensions in input matrix')
        return -1

    qout = np.zeros((dims[0], 4))

    arg = 1.0 + mi[:, 0, 0] + mi[:, 1, 1] + mi[:, 2, 2]

    idx = np.argwhere(arg < 0.0)

    if len(idx) != 0:
        arg[idx] = 0.0

    qout[:, 0] = 0.5 * np.sqrt(arg)

    arg = 1.0 + mi[:, 0, 0] - mi[:, 1, 1] - mi[:, 2, 2]

    idx = np.argwhere(arg < 0.0)

    if len(idx) != 0:
        arg[idx] = 0.0

    qout[:, 1] = 0.5 * np.sqrt(arg)

    arg = 1.0 - mi[:, 0, 0] + mi[:, 1, 1] - mi[:, 2, 2]

    idx = np.argwhere(arg < 0.0)

    if len(idx) != 0:
        arg[idx] = 0.0

    qout[:, 2] = 0.5 * np.sqrt(arg)

    arg = 1.0 - mi[:, 0, 0] - mi[:, 1, 1] + mi[:, 2, 2]

    idx = np.argwhere(arg < 0.0)

    if len(idx) != 0:
        arg[idx] = 0.0

    qout[:, 3] = 0.5 * np.sqrt(arg)

    imax = np.zeros(dims[0], dtype=int)
    dmax = np.zeros(dims[0])

    for i in range(4):
        idx = np.argwhere(np.abs(qout[:, i]) > dmax)
        if len(idx) != 0:
            imax[idx] = i
            dmax[idx] = qout[idx, i]

    idx = np.argwhere(imax == 0)

    if len(idx) != 0:
        qout[idx, 1] = (mi[idx, 2, 1] - mi[idx, 1, 2]) / (4 * qout[idx, 0])
        qout[idx, 2] = (mi[idx, 0, 2] - mi[idx, 2, 0]) / (4 * qout[idx, 0])
        qout[idx, 3] = (mi[idx, 1, 0] - mi[idx, 0, 1]) / (4 * qout[idx, 0])

    idx = np.argwhere(imax == 1)

    if len(idx) != 0:
        qout[idx, 2] = (mi[idx, 1, 0] + mi[idx, 0, 1]) / (4 * qout[idx, 1])
        qout[idx, 3] = (mi[idx, 2, 0] + mi[idx, 0, 2]) / (4 * qout[idx, 1])
        qout[idx, 0] = (mi[idx, 2, 1] - mi[idx, 1, 2]) / (4 * qout[idx, 1])

    idx = np.argwhere(imax == 2)

    if len(idx) != 0:
        qout[idx, 3] = (m[idx, 2, 1] + m[idx, 1, 2]) / (4 * qout[idx, 2])
        qout[idx, 0] = (m[idx, 0, 2] - m[idx, 2, 0]) / (4 * qout[idx, 2])
        qout[idx, 1] = (m[idx, 1, 0] + m[idx, 0, 1]) / (4 * qout[idx, 2])

    idx = np.argwhere(imax == 3)

    if len(idx) != 0:
        qout[idx, 0] = (mi[idx, 1, 0] - mi[idx, 0, 1]) / (4 * qout[idx, 3])
        qout[idx, 1] = (mi[idx, 2, 0] + mi[idx, 0, 2]) / (4 * qout[idx, 3])
        qout[idx, 2] = (mi[idx, 2, 1] + mi[idx, 1, 2]) / (4 * qout[idx, 3])

    idx = np.argwhere(qout[:, 0] < 0.0)

    if len(idx) != 0:
        qout[idx, :] = -qout[idx, :]

    qret = qnormalize(qout)

    return np.reshape(qret, (dims[0], 4))


def qtom(qi):
    """
    Purpose: transforms quaternions into rotation matrices
    WARNING!!! It appears that this routine returns the transpose (inverse) of the rotation matrix!
    It differs from the CSPICE library and Wikipedia

    Inputs: a 4 element array representing a quaternion or an Nx4 element
    array representing an array of quaternions

    Returns: a 3x3 matrix or an Nx3x3 array

    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                     q[1] = vector x
                     q[2] = vector y
                     q[3] = vector z

    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """

    if isinstance(qi, int):
        return -1

    e00 = qi[:, 0] * qi[:, 0]
    e11 = qi[:, 1] * qi[:, 1]
    e22 = qi[:, 2] * qi[:, 2]
    e33 = qi[:, 3] * qi[:, 3]
    e01 = 2 * qi[:, 0] * qi[:, 1]
    e02 = 2 * qi[:, 0] * qi[:, 2]
    e03 = 2 * qi[:, 0] * qi[:, 3]
    e12 = 2 * qi[:, 1] * qi[:, 2]
    e13 = 2 * qi[:, 1] * qi[:, 3]
    e23 = 2 * qi[:, 2] * qi[:, 3]

    mout = np.zeros((len(e00), 3, 3))

    mout[:, 0, 0] = e00 + e11 - e22 - e33
    mout[:, 1, 0] = e12 + e03
    mout[:, 2, 0] = e13 - e02
    mout[:, 0, 1] = e12 - e03
    mout[:, 1, 1] = e00 - e11 + e22 - e33
    mout[:, 2, 1] = e23 + e01
    mout[:, 1, 2] = e23 - e01
    mout[:, 0, 2] = e13 + e02
    mout[:, 2, 2] = e00 - e11 - e22 + e33

    return mout


def qcompose(vec, theta, free=True):
    """
    Purpose: compose quaternions from vectors and angles
    Inputs: vec: 3 element array or an Nx3 element array
            theta: an angle or an N element array of angles(in radians)
    Keywords: free: Flag to allow thetas outside [0,pi)
    Returns: a 4 element quaternion or an Nx4 element array of quaternions
    Notes: Implementation largely copied from the euve c library for
    quaternions
    Represention has q[0] = scalar component
                 q[1] = vector x
                 q[2] = vector y
                 q[3] = vector z
    The vector component of the quaternion can also be thought of as
    an eigenvalue of the rotation the quaterion performs

    This routine is based on the IDL version by Patrick Cruce
    """

    # Constant indicating where sin(theta) is close enough to theta
    epsilon = 1.0e-20

    vi = deepcopy(vec)
    thi = deepcopy(theta)

    # this next block of code moves angles into the range [0,PI)
    if not free:
        thi = thi % np.pi

        idx = np.argwhere(thi < 0)

        if len(idx) > 1:
            thi[idx] += np.pi

    # calculate the vector norm
    norm = np.sqrt(np.nansum(vi*vi, axis=1))

    # decide which quaternions become identity vectors
    idx1 = np.argwhere(norm < epsilon).flatten()
    idx2 = np.argwhere(norm >= epsilon).flatten()

    out_arr = np.zeros((len(norm), 4))

    if len(idx1) > 0:
        out_arr[idx1, 0] = 1.0
        out_arr[idx1, 1:4] = 0.0

    if len(idx2) > 0:
        out_arr[idx2, 0] = np.cos(thi[idx2]/2.0)

        stheta2 = np.sin(thi[idx2]/2.0)

        out_arr[idx2, 1] = (stheta2 * vi[idx2, 0])/norm[idx2]
        out_arr[idx2, 2] = (stheta2 * vi[idx2, 1])/norm[idx2]
        out_arr[idx2, 3] = (stheta2 * vi[idx2, 2])/norm[idx2]

    return out_arr



import cdflib
import numpy as np

from pytplot import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load


def camera_omti_asi(
    trange=['2020-08-01', '2020-08-02'],
    suffix='',
    site='all',
    wavelength=[5577],
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True
):

    site_code_all = ['abu', 'ath', 'drw', 'eur', 'gak', 'hlk',
                     'hus', 'isg', 'ist', 'ith', 'kap', 'ktb',
                     'mgd', 'nai', 'nyr', 'ptk', 'rik', 'rsb',
                     'sgk', 'sta', 'syo', 'trs', 'yng']

    if isinstance(wavelength, str):
        wavelengthc = wavelength.split(' ')
    elif isinstance(wavelength, int):
        wavelengthc = [str(wavelength)]
    elif isinstance(wavelength, list):
        wavelengthc = []
        for i in range(len(wavelength)):
            wavelengthc.append(str(wavelength[i]))

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = site_code_all
    
    site_code = list(set(site_code).intersection(site_code_all))

    
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        for wavelength_in in wavelengthc:
            prefix = 'omti_asi_'+site_input+'_'+wavelength_in+'_'
            file_res = 3600.
            pathformat = 'ground/camera/omti/asi/'+site_input\
                            +'/%Y/%m/%d/omti_asi_c??_'+site_input+'_'+wavelength_in+'_%Y%m%d%H_v??.cdf'

            loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                            varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
            
            if notplot:
                loaded_data.update(loaded_data_temp)
            else:
                loaded_data += loaded_data_temp
            if (len(loaded_data_temp) > 0) and ror:
                try:
                    if isinstance(loaded_data_temp, list):
                        if downloadonly:
                            cdf_file = cdflib.CDF(loaded_data_temp[-1])
                            gatt = cdf_file.globalattsget()
                        else:
                            gatt = get_data(loaded_data_temp[-1], metadata=True)['CDF']['GATT']
                    elif isinstance(loaded_data_temp, dict):
                        gatt = loaded_data_temp[list(loaded_data_temp.keys())[-1]]['CDF']['GATT']
                    print('**************************************************************************')
                    print(gatt["Logical_source_description"])
                    print('')
                    print(f'Information about {gatt["Station_code"]}')
                    print(f'PI: {gatt["PI_name"]}')
                    print('')
                    print(f'Affiliations: {gatt["PI_affiliation"]}')
                    print('')
                    print('Rules of the Road for OMTI ASI Data Use:')
                    for gatt_text in gatt["TEXT"]:
                        print(gatt_text)
                    print(f'{gatt["LINK_TEXT"]}')
                    print('**************************************************************************')
                except:
                    print('printing PI info and rules of the road was failed')
                
            if (not downloadonly) and (not notplot):
                current_tplot_name = prefix+'cloud' + suffix
                if current_tplot_name in loaded_data:
                    get_data_vars = get_data(current_tplot_name)
                    if get_data_vars is None:
                        store_data(current_tplot_name, delete=True)
                    else:
                        new_tplot_name = 'omti_asi_'+site_input+'_cloud'+suffix
                        store_data(current_tplot_name, newname=new_tplot_name)
                        loaded_data.remove(current_tplot_name)
                        if new_tplot_name not in loaded_data:
                            loaded_data.append(new_tplot_name)
                        #;--- Missing data -1.e+31 --> NaN
                        clip(new_tplot_name, -1, 9)
                    
                current_tplot_name = prefix+'image_raw' + suffix
                if current_tplot_name in loaded_data:
                    get_data_vars = get_data(current_tplot_name)
                    if get_data_vars is None:
                        store_data(current_tplot_name, delete=True)
                    else:
                        #;--- Missing data -1.e+31 --> NaN
                        clip(current_tplot_name, -1e+6, 1e+6)
                        get_data_vars = get_data(current_tplot_name)
                        """
                        Transpose y element of the image data.
                        In order to not to make an Upside down, left and right upside down,
                        for saving figure, by PIL library.
                        """
                        image_y_transpose = get_data_vars[1].transpose(0, 2, 1)
                        """
                        Try to get 'Data_Type_Description' from CDF file.
                        In order to adjust the data type of y element into original one.
                        (After clip, data type of y element may become float.)
                        This may be need for saving figure correctly.
                        """
                        file_name = get_data(current_tplot_name,
                                            metadata=True)['CDF']['FILENAME']
                        if isinstance(file_name, list):
                            if len(file_name) > 0:
                                file_name = file_name[0]
                        cdf_file = cdflib.CDF(file_name)
                        cdf_info = cdf_file.cdf_info()
                        all_cdf_variables = cdf_info['rVariables'] + cdf_info['zVariables']
                        if 'image_raw' in all_cdf_variables:
                            var_string = 'image_raw'
                            var_properties = cdf_file.varinq(var_string)
                            if 'Data_Type_Description' in var_properties:
                                original_datatype_string = var_properties['Data_Type_Description']
                                if original_datatype_string == 'CDF_INT4':
                                    image_y_transpose = image_y_transpose.astype(np.int32)
                                elif original_datatype_string == 'CDF_UINT1':
                                    image_y_transpose = image_y_transpose.astype(np.uint8)
                                elif original_datatype_string == 'CDF_UINT2':
                                    image_y_transpose = image_y_transpose.astype(np.uint16)
                                elif original_datatype_string == 'CDF_UINT4':
                                    image_y_transpose = image_y_transpose.astype(np.uint32)

                        get_metadata_vars = get_data(current_tplot_name, metadata=True)
                        store_data(current_tplot_name,
                                   data={'x':get_data_vars[0],
                                         'y':image_y_transpose},
                                   attr_dict=get_metadata_vars)

    return loaded_data



"""
Performs coordinate transformations for MMS data using MMS MEC quaternions.

This function uses the mms_cotrans_qtransformer function to recursively transform the input data from the input coordinate system to the output coordinate system by going through ECI. The transformation operations are performed using the SpacePy library.

Parameters:
in_name (str or list of str): Names of Tplot variables of vectors to be transformed.
out_name (str or list of str): Names of output variables.
in_coord (str): Input coordinate system (e.g. 'bcs', 'gse', 'gse2000', 'gsm', 'sm', 'geo', 'eci').
out_coord (str): Output coordinate system (e.g. 'bcs', 'gse', 'gse2000', 'gsm', 'sm', 'geo', 'eci').
probe (str): MMS spacecraft # (must be '1', '2', '3', or '4').

Returns:
list of str: List of variables created.
"""
import logging
from pyspedas import cotrans_get_coord
from pytplot import get_data
from .mms_cotrans_qtransformer import mms_cotrans_qtransformer

try:
    import spacepy.coordinates as coord
except ImportError:
    logging.error('SpacePy must be installed to use this module.')
    logging.error('Please install it using: pip install spacepy')


def mms_qcotrans(in_name=None, out_name=None, in_coord=None, out_coord=None, probe=None):
    """
    Perform coordinate transformations using MMS MEC quaternions.

    This routine uses SpacePy to do the quaternion transformation operations.

    Parameters
    -----------
        in_name: str or list of str
            Names of Tplot variables of vectors to be transformed

        out_name: str or list of str
            Names of output variables

        in_coord: str
            Input coordinate system (e.g. 'bcs','gse','gse2000','gsm','sm','geo','eci')

        out_coord: str
            Output coordinate system (e.g. 'bcs','gse','gse2000','gsm','sm','geo','eci')

        probe: str
            MMS spacecraft # (must be 1, 2, 3, or 4)

    Returns
    --------
        List of variables created

    """
    valid_probes = ['1', '2', '3', '4']
    valid_coords = ['bcs', 'dbcs', 'dmpa', 'smpa', 'dsl', 'ssl', 'gse', 'gse2000', 'gsm', 'sm', 'geo', 'eci', 'j2000']

    if in_name is None:
        logging.error('Input variable name is missing')
        return

    if out_name is None:
        logging.error('Output variable name is missing')
        return

    if not isinstance(in_name, list):
        in_name = [in_name]

    if not isinstance(out_name, list):
        out_name = [out_name]

    out_vars = []

    for idx, variable in enumerate(in_name):
        if in_coord is None:
            var_coords = cotrans_get_coord(variable)
            in_coord = var_coords
            if var_coords is None:
                logging.error('Could not determine coordinate system for: ' + variable)
                continue

        if isinstance(in_coord, list):
            var_coords = in_coord[idx]
        else:
            var_coords = in_coord

        var_coords = var_coords.lower()

        if var_coords not in valid_coords:
            logging.error('Unsupported input coordinate system: ' + var_coords)
            continue

        if isinstance(out_coord, list):
            new_coords = out_coord[idx]
        else:
            new_coords = out_coord

        new_coords = new_coords.lower()

        if new_coords not in valid_coords:
            logging.error('Unsupported output coordinate system: ' + new_coords)

        if var_coords in ['bcs', 'ssl'] or new_coords in ['bcs', 'ssl']:
            logging.warning('WARNING: there are issues transforming data to/from a spinning coordinate system')

        # find the probe, if it's not specified by the user
        if probe is None:
            name_pieces = variable.split('_')
            if len(name_pieces) <= 1:
                logging.warning('Probe could not be determined from: ' + variable + '; defaulting to probe 1')
                probe = '1'
            else:
                probe = name_pieces[0][-1]

        if not isinstance(probe, str):
            probe = str(probe)

        if probe not in valid_probes:
            logging.error('Unknown probe for variable: ' + variable + '; continuing without transforming...')
            continue

        transformed = mms_cotrans_qtransformer(variable, out_name[idx], var_coords, new_coords, probe=probe)

        if transformed is not None:
            out_vars.append(transformed)

            # update the coordinate system in the plot metadata if needed
            metadata = get_data(transformed, metadata=True)
            if metadata.get('plot_options') is not None:
                if metadata['plot_options'].get('yaxis_opt') is not None:
                    if metadata['plot_options']['yaxis_opt'].get('legend_names') is not None:
                        legend = metadata['plot_options']['yaxis_opt'].get('legend_names')
                        updated_legend = [item.replace(var_coords.upper(), new_coords.upper()) for item in legend]
                        metadata['plot_options']['yaxis_opt']['legend_names'] = updated_legend
                    if metadata['plot_options']['yaxis_opt'].get('axis_label') is not None:
                        ytitle = metadata['plot_options']['yaxis_opt'].get('axis_label')
                        metadata['plot_options']['yaxis_opt']['axis_label'] = ytitle.replace(var_coords.upper(),
                                                                                             new_coords.upper())
    return out_vars



import logging
import os
import re
import shutil
from .mms_config import CONFIG
from .mms_files_in_interval import mms_files_in_interval
from dateutil.rrule import rrule, DAILY
from dateutil.parser import parse
from datetime import timedelta


def mms_get_local_files(probe, instrument, data_rate, level, datatype, trange, mirror=False):
    """
    Search for local MMS files in case a list cannot be retrieved from the
    remote server.  
    
    Parameters
    ------------
        probe: str
            probe #, e.g., '4' for MMS4

        instrument: str
            instrument name, e.g., 'fpi' or 'fgm'

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level, e.g., 'l2'

        datatype: str
            'electron' or 'ion'

        trange: list of str
            two-element array containing the start and end date/times

        mirror: bool
            if True, copy files from network mirror to local data directory

    Returns
    ---------
        List of file paths.
    """
    files_out = []

    if mirror:
        if CONFIG.get('mirror_data_dir') is not None:
            data_dir = CONFIG['mirror_data_dir']
        else:
            return []
    else:
        data_dir = CONFIG['local_data_dir']

    # directory and file name search patterns
    #   -assume directories are of the form:
    #      (srvy, SITL): spacecraft/instrument/rate/level[/datatype]/year/month/
    #      (brst): spacecraft/instrument/rate/level[/datatype]/year/month/day/
    #   -assume file names are of the form:
    #      spacecraft_instrument_rate_level[_datatype]_YYYYMMDD[hhmmss]_version.cdf

    file_name = 'mms'+probe+'_'+instrument+'_'+data_rate+'_'+level+'(_)?.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf'

    days = rrule(DAILY, dtstart=parse(parse(trange[0]).strftime('%Y-%m-%d')), until=parse(trange[1])-timedelta(seconds=1))

    if datatype == '' or datatype is None:
        level_and_dtype = level
    else:
        level_and_dtype = os.sep.join([level, datatype])

    for date in days:
        if data_rate == 'brst':
            local_dir = os.sep.join([data_dir, 'mms'+probe, instrument, data_rate, level_and_dtype, date.strftime('%Y'), date.strftime('%m'), date.strftime('%d')])
        else:
            local_dir = os.sep.join([data_dir, 'mms'+probe, instrument, data_rate, level_and_dtype, date.strftime('%Y'), date.strftime('%m')])

        if os.name == 'nt':
            full_path = os.sep.join([re.escape(local_dir)+os.sep, file_name])
        else:
            full_path = os.sep.join([re.escape(local_dir), file_name])

        # check for extra /'s in the path
        if '//' in full_path:
            full_path = full_path.replace('//', '/')

        regex = re.compile(full_path)
        for root, dirs, files in os.walk(data_dir):
            for file in files:
                this_file = os.sep.join([root, file])
                if CONFIG['debug_mode']: logging.info('Checking ' + this_file)
                if CONFIG['debug_mode']: logging.info('against: ' + full_path)
                
                matches = regex.match(this_file)
                if matches:
                    this_time = parse(matches.groups()[1])
                    if this_time >= parse(parse(trange[0]).strftime('%Y-%m-%d')) and this_time <= parse(trange[1])-timedelta(seconds=1):
                        if not any(this_file == f['full_name'] for f in files_out):
                            files_out.append({'file_name': file, 'timetag': '', 'full_name': this_file, 'file_size': ''})

    files_in_interval = mms_files_in_interval(files_out, trange)

    local_files = []

    file_names = [f['file_name'] for f in files_in_interval]

    for file in files_out:
        if file['file_name'] in file_names:
            local_files.append(file['full_name'])

    if mirror:
        mirror_dir = CONFIG['mirror_data_dir']
        local_dir = CONFIG['local_data_dir']
        local_files_copied = []
        # need to copy files from network mirror to local data directory
        for file in local_files:
            local_file = file.replace(mirror_dir, local_dir)
            if CONFIG['debug_mode']:
                logging.info('Copying ' + file + ' to ' + local_file)
            shutil.copyfile(file, local_file)
            local_files_copied.append(local_file)
        local_files = local_files_copied

    for file in local_files:
        logging.info('Loading: ' + file)

    return local_files



"""
This function transforms MMS vector fields from GSM coordinates to LMN (boundary-normal) coordinates using the Shue et al., 1998 magnetopause model. The input and output tplot variables are specified by name_in and name_out, respectively. Additional optional parameters include specifying the input data coordinates (GSM or GSE), probe, and data rate. The function returns the name of the output variable containing the data in LMN coordinates.
"""

import numpy as np
import logging
from pytplot import get_data, store_data, options
from pyspedas.cotrans.cotrans_get_coord import cotrans_get_coord
from pyspedas.cotrans.cotrans import cotrans
from pyspedas.cotrans.gsm2lmn import gsm2lmn
from pyspedas.mms import mec
from pyspedas import tinterpol, omni

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_cotrans_lmn(name_in, name_out, gsm=False, gse=False, probe=None, data_rate='srvy'):
    """
    Tranforms MMS vector fields from GSM coordinates to LMN (boundary-normal) coordinates
    using the Shue et al., 1998 magnetopause model

    Input
    ------
        name_in: str
            Name of the input tplot variable

        name_out: str
            Name of the output tplot variable

    Parameters
    -----------
        gsm: bool
            Flag to indicate the input data are in GSM coordinates

        gse: bool
            Flag to indicate the input data are in GSE coordinates; note: the input data 
            will be transformed to GSM prior to transforming to LMN

        probe: str
            Spacecraft probe #; if not specified, the routine attempts to extract from the 
            input variable name

        data_rate: str
            Data rate of the input data; default is 'srvy'

    Returns
    --------
        Name of the variable containing the data in LMN coordinates.

    """

    data_in = get_data(name_in)
    metadata_in = get_data(name_in, metadata=True)

    if data_in is None:
        logging.error('Error reading tplot variable: ' + name_in)
        return None

    data_in_coord = cotrans_get_coord(name_in).lower()

    if data_in_coord != 'gse' and data_in_coord != 'gsm' and not gse and not gsm:
        logging.error('Please specify the coordinate system of the input data.')
        return

    # we'll need the probe if it's not specified via keyword
    if probe is None:
        sc_id = name_in.split('_')[0]

        if sc_id != '':
            probe = sc_id[-1]

        if probe not in ['1', '2', '3', '4']:
            logging.error('Error, probe not found; please specify the probe via the "probe" keyword.')
            return

    # load the spacecraft position data
    mec_vars = mec(trange=[min(data_in.times), max(data_in.times)], probe=probe, data_rate=data_rate)

    # interpolate the position data to the input data
    tinterp_vars = tinterpol('mms'+probe+'_mec_r_gsm', name_in, newname='mms'+probe+'_mec_r_gsm_interp')

    pos_data = get_data('mms'+probe+'_mec_r_gsm_interp')

    # we'll need the input data in GSM coordinates
    if data_in_coord != 'gsm':
        cotrans_out = cotrans(name_in, name_in + '_gsm', coord_out='gsm')
        data_in = get_data(name_in + '_gsm')
    
    swdata = solarwind_load([min(data_in.times), max(data_in.times)])

    logging.info('Transforming GSM -> LMN; this may take several minutes, depending on the size of the input.')
    Blmn = gsm2lmn(data_in.times, pos_data.y, data_in.y, swdata=swdata)

    saved = store_data(name_out, data={'x': data_in.times, 'y': Blmn}, attr_dict=metadata_in)

    if saved:
        options(name_out, 'legend_names', ['L', 'M', 'N'])
        return name_out
    else:
        logging.error('Problem creating tplot variable.')


def solarwind_load(trange, level='hro2', min5=False):
    """
    Loads solar wind data for use in the GSM to LMN transformation.

    Parameters
    ----------
        trange: list of float
            Time range of data to be loaded

        level: str
            Data level (default: hro2)

        min5: bool
            Flag indicating whether to load 1 minute or 5 minute data (default: 1 minute)

    Returns
    -------
        Numpy array of solar wind data with shape (N, 3), where N is the number of time points
        and the columns are the time, Bz GSM, and pressure.
    """

    if min5:
        datatype = '5min'
    else:
        datatype = '1min'
    omni_vars = omni.data(trange=trange, level=level, datatype=datatype)
    bzgsm = get_data('BZ_GSM')
    dp = get_data('Pressure')
    return np.array([bzgsm.times, bzgsm.y, dp.y]).T



import os
import unittest
from pytplot import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.twins.imager(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_img_data(self):
        img_vars = pyspedas.twins.imager()
        self.assertTrue(data_exists('smooth_image_val'))
        img_vars = pyspedas.twins.imager(notplot=True)
        self.assertTrue('smooth_image_val' in img_vars)

    def test_load_lad_data(self):
        lad_vars = pyspedas.twins.lad(time_clip=True)
        self.assertTrue(data_exists('lad1_data'))
        self.assertTrue(data_exists('lad2_data'))

    def test_load_ephem_data(self):
        ephemeris_vars = pyspedas.twins.ephemeris()
        self.assertTrue(data_exists('FLTGEO'))


if __name__ == '__main__':
    unittest.main()



"""
Time clip of data.

Notes
-----
Similar to tclip.pro in IDL SPEDAS.

"""
import logging
import pytplot


def time_clip(names, time_start, time_end, new_names=None, suffix=None,
              overwrite=None):
    """
    Clip data from time_start to time_end.

    Parameters:
    names: str/list of str
        List of pytplot names.
    time_start : float
        Start time.
    time_end : float
        End time.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If '', then pytplot variables are replaced.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-m'.
    overwrite: bool, optional
        Replace the existing tplot name.

    Returns
    -------
    None.

    """
    logging.info("time_clip has been moved to the pytplot.tplot_math module. Please update your imports!")
    logging.info("This version will eventually be removed.")
    pytplot.tplot_math.time_clip(names=names,time_start=time_start,time_end=time_end,new_names=new_names,suffix=suffix,
                                 overwrite=overwrite)


from .load import load
from pyspedas.utilities.datasets import find_datasets


def mpa(trange=['2004-10-31', '2004-11-01'],
        probe='l1', 
        level='k0', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magnetospheric Plasma Analyzer (MPA)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            LANL probe #; Valid options:
                'l0' for LANL 1990 data
                'l1' for LANL 1991 data
                'l4' for LANL 1994 data
                'l7' for LANL 1997 data
                'l9' for LANL 1989 data
                'a1' for LANL 2001 data
                'a2' for LANL 2002 data

        level: str
            Data level; options: 'k0' (default: k0)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='mpa', trange=trange, level=level, probe=probe, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mpa_postprocessing(tvars)


def mpa_postprocessing(variables):
    """
    Placeholder for MPA post-processing 
    """
    return variables


def spa(trange=['2004-10-31', '2004-11-01'],
        probe='l1', 
        level='k0', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Synchronous Orbit Particle Analyzer (SPA)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            LANL S/C probe #; Valid options:
                'l0' for LANL 1990 data
                'l1' for LANL 1991 data
                'l4' for LANL 1994 data
                'l7' for LANL 1997 data
                'l9' for LANL 1989 data
                'a1' for LANL 2001 data
                'a2' for LANL 2002 data

        level: str
            Data level; options: 'k0' (default: k0)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='spa', trange=trange, level=level, probe=probe, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return spa_postprocessing(tvars)


def spa_postprocessing(variables):
    """
    Placeholder for SPA post-processing 
    """
    return variables


def datasets(instrument=None, label=True):
    return find_datasets(mission='LANL', instrument=instrument, label=label)




import os
import unittest
from pytplot import data_exists

import pyspedas
from pytplot import del_data


class LoadTestCases(unittest.TestCase):

    def test_downloadonly(self):
        del_data()
        mag_files = pyspedas.goes.fgm(datatype='1min', downloadonly=True)
        self.assertTrue(os.path.exists(mag_files[0]))

    def test_load_orbit_data(self):
        del_data()
        orbit_vars = pyspedas.goes.orbit(downloadonly=True)
        orbit_vars = pyspedas.goes.orbit(notplot=True)
        orbit_vars = pyspedas.goes.orbit()
        self.assertTrue(data_exists('XYZ_GSM'))
        self.assertTrue(data_exists('XYZ_GSE'))
        self.assertTrue(data_exists('XYZ_SM'))
        self.assertTrue(data_exists('XYZ_GEO'))

    def test_load_1min_mag_data(self):
        del_data()
        mag_vars = pyspedas.goes.fgm(datatype='1min')
        self.assertTrue(data_exists('BX_1'))
        self.assertTrue(data_exists('BY_1'))
        self.assertTrue(data_exists('BZ_1'))

    def test_load_5min_mag_data(self):
        del_data()
        mag_vars = pyspedas.goes.fgm(datatype='5min', probe='10', trange=['2000-07-01', '2000-07-02'], time_clip=True)
        self.assertTrue(data_exists('ht'))

    def test_load_full_mag_data(self):
        del_data()
        mag_vars = pyspedas.goes.fgm(datatype='512ms', suffix='_512')
        self.assertTrue(data_exists('BX_1_512'))
        self.assertTrue(data_exists('BY_1_512'))
        self.assertTrue(data_exists('BZ_1_512'))

    def test_load_1min_epead_data(self):
        del_data()
        epead_vars = pyspedas.goes.epead()
        self.assertTrue(data_exists('E1E_UNCOR_FLUX'))
        self.assertTrue(data_exists('E1W_UNCOR_FLUX'))
        self.assertTrue(data_exists('E2E_UNCOR_FLUX'))

    def test_load_full_epead_data(self):
        del_data()
        epead_vars = pyspedas.goes.epead(datatype='1min')
        self.assertTrue(data_exists('E1E_UNCOR_FLUX'))
        self.assertTrue(data_exists('E1W_UNCOR_FLUX'))
        self.assertTrue(data_exists('E2E_UNCOR_FLUX'))

    def test_load_5min_epead_data(self):
        del_data()
        epead_vars = pyspedas.goes.epead(datatype='5min')
        self.assertTrue(data_exists('E1E_UNCOR_FLUX'))
        self.assertTrue(data_exists('E1W_UNCOR_FLUX'))
        self.assertTrue(data_exists('E2E_UNCOR_FLUX'))

    def test_load_full_maged_data(self):
        del_data()
        maged_vars = pyspedas.goes.maged(datatype='full')
        self.assertTrue(data_exists('M_1ME1_DTC_UNCOR_CR'))

    def test_load_1min_maged_data(self):
        del_data()
        maged_vars = pyspedas.goes.maged(datatype='1min', time_clip=True)
        self.assertTrue(data_exists('M_1ME1_DTC_UNCOR_FLUX'))

    def test_load_5min_maged_data(self):
        del_data()
        maged_vars = pyspedas.goes.maged(datatype='5min', time_clip=True)
        self.assertTrue(data_exists('M_2ME1_DTC_COR_FLUX'))

    def test_load_full_magpd_data(self):
        del_data()
        magpd_vars = pyspedas.goes.magpd(datatype='full')
        self.assertTrue(data_exists('M_1MP1_DTC_UNCOR_CR'))

    def test_load_1min_magpd_data(self):
        del_data()
        magpd_vars = pyspedas.goes.magpd(datatype='1min', time_clip=True)
        self.assertTrue(data_exists('M_1MP1_DTC_UNCOR_FLUX'))

    def test_load_full_hepad_data(self):
        del_data()
        hepad_vars = pyspedas.goes.hepad(datatype='full')
        self.assertTrue(data_exists('P10_FLUX'))

    def test_load_1min_hepad_data(self):
        del_data()
        hepad_vars = pyspedas.goes.hepad(prefix='probename', time_clip=True)
        self.assertTrue(data_exists('g15_P10_FLUX'))

    def test_load_xrs_data(self):
        del_data()
        xrs_vars = pyspedas.goes.xrs(probe='10', datatype='full', trange=['2002-08-01', '2002-08-01'])
        self.assertTrue(data_exists('xl'))

    def test_load_xrs_5m_data(self):
        del_data()
        xrs_vars = pyspedas.goes.xrs(probe='11', datatype='5min', trange=['2000-09-01', '2000-09-01'], time_clip=True)
        self.assertTrue(data_exists('xl'))

    def test_load_xrs_1m_data(self):
        del_data()
        xrs_vars = pyspedas.goes.xrs(probe='11', datatype='1min', trange=['2000-09-01', '2000-09-01'], prefix='probename', time_clip=True)
        self.assertTrue(data_exists('g11_xs'))

    def test_load_eps_1m_data(self):
        del_data()
        eps_vars = pyspedas.goes.eps(trange=['2000-09-01', '2000-09-01'], probe='11', time_clip=True)
        self.assertTrue(data_exists('e1_flux_i'))
        self.assertTrue(data_exists('e2_flux_i'))
        self.assertTrue(data_exists('e3_flux_i'))

    def test_load_eps_5m_data(self):
        del_data()
        eps_vars = pyspedas.goes.eps(trange=['2000-09-01', '2000-09-01'], probe='11', datatype='5min', time_clip=True)
        self.assertTrue(data_exists('p1_flux'))
        self.assertTrue(data_exists('p2_flux'))
        self.assertTrue(data_exists('p3_flux'))
        self.assertTrue(data_exists('p4_flux'))
        self.assertTrue(data_exists('p5_flux'))
        self.assertTrue(data_exists('p6_flux'))
        self.assertTrue(data_exists('p7_flux'))

    def test_load_xrs_data_16(self):
        del_data()
        xrs_vars_16 = pyspedas.goes.xrs(probe='16', trange=['2022-09-01', '2022-09-02'])
        self.assertTrue('xrsa_flux' in xrs_vars_16)

    def test_load_xrs_data_17_hi(self):
        del_data()
        xrs_vars_17 = pyspedas.goes.xrs(probe='17', trange=['2022-07-01', '2022-07-02'], datatype='hi')
        self.assertTrue('xrsb_flux' in xrs_vars_17)

    def test_load_euvs_data_17(self):
        del_data()
        euvs_vars_17 = pyspedas.goes.euvs(probe='17', trange=['2022-09-01', '2022-09-02'], prefix='probename')
        self.assertTrue('g17_irr_256' in euvs_vars_17)

    def test_load_euvs_data_16_hi(self):
        del_data()
        euvs_vars_16 = pyspedas.goes.euvs(probe='16', trange=['2022-08-01', '2022-08-02'], prefix='probename', datatype='hi')
        self.assertTrue('g16_irr_256' in euvs_vars_16)

    def test_load_mag_data_16(self):
        del_data()
        mag_vars_16 = pyspedas.goes.mag(probe='16', trange=['2023-01-30', '2023-01-31'], prefix='goes16_')
        self.assertTrue('goes16_b_total' in mag_vars_16)

    def test_load_mag_data_17(self):
        del_data()
        mag_vars_17 = pyspedas.goes.mag(probe='17', trange=['2022-01-30', '2022-01-31'], prefix='goes17_', datatype='hi')
        self.assertTrue('goes17_b_gse' in mag_vars_17)

    def test_load_mpsh_data_17(self):
        del_data()
        mpsh_vars_17 = pyspedas.goes.mpsh(probe='17', trange=['2022-09-01', '2022-09-02'], prefix='probename', time_clip=True)
        self.assertTrue('g17_AvgDiffElectronFlux' in mpsh_vars_17)

    def test_load_mpsh_data_16(self):
        del_data()
        mpsh_vars_16 = pyspedas.goes.mpsh(probe='16', trange=['2022-10-01', '2022-10-02'], prefix='probename', datatype='hi')
        self.assertTrue('g16_AvgDiffProtonFlux' in mpsh_vars_16)

    def test_load_sgps_data_16(self):
        del_data()
        sgps_vars_16 = pyspedas.goes.sgps(probe='16', trange=['2023-01-30', '2023-01-31'], prefix='probename')
        self.assertTrue('g16_AvgDiffAlphaFlux' in sgps_vars_16)

    def test_load_sgps_data_18(self):
        del_data()
        sgps_vars_18 = pyspedas.goes.sgps(probe='18', trange=['2023-01-30', '2023-01-31'], prefix='probename', datatype='hi', time_clip=True)
        self.assertTrue('g18_AvgIntProtonFlux' in sgps_vars_18)


if __name__ == '__main__':
    unittest.main()



import numpy as np


def slice2d_rlog(r, dr):
    """
    Apply radial log scaling to aggregated velocity/energy vectors.

    Input
    ------
        r: ndarray
            N element array of radii

        dr: ndarray
            N element array of radial bin widths (full width)

    Returns
    --------
        Returns hash table containing scaled radii and bin widths
    """

    # get log of radial boundaries
    rbottom = np.log10(r - 0.5*dr)
    rtop = np.log10(r + 0.5*dr)

    rrange = [np.nanmin(rbottom), np.nanmax(rtop)]
    span = rrange[1] - rrange[0]

    # shrink gap between (0, r_min] and normalize
    rbottom = rbottom - rrange[0]
    rbottom = rbottom/span

    rtop = rtop - rrange[0]
    rtop = rtop/span

    r = (rtop + rbottom)/2.0
    dr = (rtop - rbottom)

    return {'r': r, 'dr': dr}


import logging
import warnings
import numpy as np
from pyspedas import tnames
from pytplot import get_data, store_data, options

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_eis_pad_spinavg(scopes=['0','1','2','3','4','5'], probe='1', 
    data_rate='srvy', level='l2', datatype='extof', data_units='flux', 
    species='proton', energy=[55, 800], size_pabin=15, suffix=''):
    """
    Calculate spin-averaged pitch angle distributions using data from the MMS Energetic Ion Spectrometer (EIS)
    
    Parameters
    ----------
        scopes: list of str
            telescope #s to include in the calculation

        probe: str
            probe #, e.g., '4' for MMS4

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst' (default: 'srvy')

        level: str
            data level ['l1a','l1b','l2pre','l2' (default)]

        datatype: str
            'extof' or 'phxtof' (default: 'extof')

        data_units: str
            'flux' or 'cps' (default: 'flux')

        species: str
            species for calculation (default: 'proton')

        energy: list of float
            energy range to include in the calculation (default: [55, 800])

        size_pabin: int
            size of the pitch angle bins, in degrees (default: 15)

        suffix: str
            suffix of the loaded data

    Returns
    --------
        Name of tplot variables created.
    """

    if data_units == 'cps':
        units_label = '1/s'
    else:
        units_label = '1/(cm^2-sr-s-keV)'

    if len(scopes) == 1:
        scope_suffix = '_t' + scopes + suffix
    elif len(scopes) == 6:
        scope_suffix = '_omni' + suffix

    prefix = 'mms' + probe + '_epd_eis_' + data_rate + '_' + level + '_'

    # get the spin #s associated with each measurement
    spin_times, spin_nums = get_data(prefix + datatype + '_spin' + suffix)

    # find where the spins start
    spin_starts = [spin_start for spin_start in np.where(spin_nums[1:] > spin_nums[:-1])[0]]

    pad_vars = tnames(prefix + datatype + '_*keV_' + species + '_' + data_units + scope_suffix + '_pad')

    out_vars = []

    for pad_var in pad_vars:
        pad_times, pad_data, pad_angles = get_data(pad_var)

        if pad_data is None:
            logging.error('Error, variable containing valid PAD data missing: ' + pad_var)
            continue

        spin_avg_flux = np.zeros([len(spin_starts), len(pad_angles)])
        spin_times = np.zeros(len(spin_starts))

        current_start = 0

        # loop through the spins for this telescope
        for spin_idx, spin_start in enumerate(spin_starts):
            ind = np.where(spin_nums == spin_nums[spin_starts[spin_idx]])[0]
            if ind is not None:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    spin_avg_flux[spin_idx, :] = nanmean(pad_data[ind], axis=0)
            spin_times[spin_idx] = pad_times[current_start]
            current_start = spin_starts[spin_idx]+1

        newname = pad_var + '_spin'

        n_pabins = 180./size_pabin
        new_pa_label = [180.*n_pabin/n_pabins+size_pabin/2. for n_pabin in range(0, int(n_pabins))]

        store_data(newname, data={'x': spin_times, 'y': spin_avg_flux, 'v': new_pa_label})
        options(newname, 'ylog', False)
        options(newname, 'zlog', True)
        options(newname, 'spec', True)
        options(newname, 'ztitle', units_label)
        options(newname, 'ytitle', 'MMS' + str(probe) + ' ' + datatype + ' spin PAD')
        options(newname, 'ysubtitle', '[deg]')
        out_vars.append(newname)

    return out_vars



import os
import logging
import unittest
from pytplot import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.wind.mfi(trange=['2013-11-5', '2013-11-6'], downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_3dp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2003-09-5', '2003-09-6'], notplot=True)
        self.assertTrue('N_e_dens_wi_3dp' in tdp_vars)

    def test_load_3dp_pm_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2018-11-06', '2018-11-07'], datatype='3dp_pm', time_clip=True, notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_pm_P_VELS' in tdp_vars)

    def test_load_3dp_ehpd_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2018-11-06', '2018-11-07'], datatype='3dp_ehpd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_ehpd_FLUX' in tdp_vars)

    def test_load_3dp_ehsp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_ehsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_ehsp_FLUX' in tdp_vars)

    def test_load_3dp_elm2_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_elm2', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_elm2_FLUX' in tdp_vars)

    def test_load_3dp_elpd_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_elpd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_elpd_FLUX' in tdp_vars)

    def test_load_3dp_elsp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_elsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_elsp_FLUX' in tdp_vars)

    def test_load_3dp_em_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_em', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_em_E_DENS' in tdp_vars)

    def test_load_3dp_emfits_e0_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2003-10-11', '2003-10-12'], datatype='3dp_emfits_e0', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('N_e_dens_wi_3dp' in tdp_vars)

    def test_load_3dp_k0_data(self):
        # Note: this datatype is periodically pruned, SPDF seems to keep the last 3 years worth.
        tdp_vars = pyspedas.wind.threedp(trange=['2023-01-01', '2023-01-02'], datatype='3dp_k0', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_k0_elect_density' in tdp_vars)

    def test_load_3dp_phsp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_phsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_phsp_FLUX' in tdp_vars)
    def test_varformat_star(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_phsp', time_clip=True,
                                         notplot=True, addmaster=True, varformat='*')
        self.assertTrue('wi_3dp_phsp_FLUX' in tdp_vars)

    def test_varformat_dots(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_phsp', time_clip=True,
                                         notplot=True, addmaster=True, varformat='....')
        self.assertTrue('wi_3dp_phsp_FLUX' in tdp_vars)

    def test_load_3dp_plsp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_plsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_plsp_FLUX' in tdp_vars)

    def test_load_3dp_varformat_alternation(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_plsp', time_clip=True,
                                         notplot=True, addmaster=True,varformat='MOM\.P\.FLUX|MOM\.P\.VELOCITY|MOM\.P\.PTENS')
        self.assertTrue('wi_3dp_plsp_MOM.P.FLUX' in tdp_vars)
        self.assertTrue('wi_3dp_plsp_MOM.P.VELOCITY' in tdp_vars)
        self.assertTrue('wi_3dp_plsp_MOM.P.PTENS' in tdp_vars)

    def test_load_3dp_sfpd_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfpd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sfpd_FLUX' in tdp_vars)
    def test_load_3dp_sfsp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sfsp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sfsp_FLUX' in tdp_vars)

    def test_load_3dp_sopd_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sopd', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sopd_FLUX' in tdp_vars)

    def test_load_3dp_sosp_data(self):
        tdp_vars = pyspedas.wind.threedp(trange=['2019-11-06', '2019-11-07'], datatype='3dp_sosp', time_clip=True,
                                         notplot=True, addmaster=True)
        self.assertTrue('wi_3dp_sosp_FLUX' in tdp_vars)

    def test_load_mfi_data(self):
        mfi_vars = pyspedas.wind.mfi(trange=['2013-11-5', '2013-11-6'], time_clip=True)
        self.assertTrue(data_exists('BGSE'))

    def test_load_swe_data(self):
        swe_vars = pyspedas.wind.swe(trange=['2013-11-5', '2013-11-6'])
        self.assertTrue(data_exists('N_elec'))
        self.assertTrue(data_exists('T_elec'))
        self.assertTrue(data_exists('W_elec'))

    def test_load_waves_data(self):
        swe_vars = pyspedas.wind.waves(trange=['2013-11-5', '2013-11-6'])
        self.assertTrue(data_exists('E_VOLTAGE_RAD1'))
        self.assertTrue(data_exists('E_VOLTAGE_RAD2'))
        self.assertTrue(data_exists('E_VOLTAGE_TNR'))

    def test_load_orbit_data(self):
        orb_vars = pyspedas.wind.orbit(trange=['2013-11-5', '2013-11-6'])
        self.assertTrue(data_exists('GSM_POS'))
        self.assertTrue(data_exists('GSM_VEL'))
        self.assertTrue(data_exists('SUN_VECTOR'))
        self.assertTrue(data_exists('GCI_POS'))
        self.assertTrue(data_exists('GCI_VEL'))

    def test_load_sms_data(self):
        sms_vars = pyspedas.wind.sms()
        self.assertTrue(data_exists('Alpha_vel'))
        self.assertTrue(data_exists('C/O_ratio'))
        self.assertTrue(data_exists('C_ion_temp'))
        self.assertTrue(data_exists('O_ion_temp'))


if __name__ == '__main__':
    unittest.main()



import cdflib
import numpy as np
from pyspedas import tnames
from pytplot import time_float
from pytplot import get_data, options, store_data, ylim, zlim

from ..load import load


def pwe_efd(trange=['2017-04-01', '2017-04-02'],
            datatype='E_spin',
            level='l2',
            suffix='',
            coord='dsi',
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            uname=None,
            passwd=None,
            time_clip=False,
            ror=True):
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        ror: bool
            If set, print PI info and rules of the road

    Returns:
        List of tplot variables created.

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True
    file_res = 3600. * 24
    prefix = 'erg_pwe_efd_'+level+'_'

    if ('64' in datatype) or ('256' in datatype):
        if '64' in datatype:
            mode = '64Hz'
        elif '256' in datatype:
            mode = '256Hz'
        md = 'E'+mode
        pathformat = 'satellite/erg/pwe/efd/'+level+'/'+md + \
            '/%Y/%m/erg_pwe_efd_'+level+'_'+md+'_'+coord+'_%Y%m%d_v??_??.cdf'
        prefix += md + '_' + coord + '_'
        if coord == 'wpt':
            component = ['Eu_waveform', 'Ev_waveform']
        elif coord == 'dsi':
            component = ['Ex_waveform', 'Ey_waveform',
                         'Eu_offset'+'_'+mode, 'Ev_offset'+'_'+mode]

    else:
        pathformat = 'satellite/erg/pwe/efd/'+level+'/'+datatype + \
            '/%Y/%m/erg_pwe_efd_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'
        prefix += datatype + '_'
        if 'spin' in datatype:
            component = ['Eu', 'Ev', 'Eu1', 'Ev1', 'Eu2', 'Ev2']
            labels = ['Ex', 'Ey']
        if datatype == 'pot':
            component = ['Vu1', 'Vu2', 'Vv1', 'Vv2']
        if datatype == 'pot8Hz':
            component = ['Vu1_waveform_8Hz', 'Vu2_waveform_8Hz',
                         'Vv1_waveform_8Hz', 'Vv2_waveform_8Hz']

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)

    if (len(loaded_data) > 0) and ror:

        try:
            if isinstance(loaded_data, list):
                if downloadonly:
                    cdf_file = cdflib.CDF(loaded_data[-1])
                    gatt = cdf_file.globalattsget()
                else:
                    gatt = get_data(loaded_data[-1], metadata=True)['CDF']['GATT']
            elif isinstance(loaded_data, dict):
                gatt = loaded_data[list(loaded_data.keys())[-1]]['CDF']['GATT']

            # --- print PI info and rules of the road

            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE EFD')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: "+gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/EFD: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Efd')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    time_min_max = time_float(trange)
    if 'spin' in datatype:
        for elem in component:
            t_plot_name = prefix + elem + '_dsi'
            options(t_plot_name, 'ytitle', elem + ' vector in DSI')
            options(t_plot_name, 'legend_names', labels)
            # ylim settings because pytplot.timespan() doesn't affect in ylim.
            # May be it will be no need in future.
            get_data_vars = get_data(t_plot_name)
            if get_data_vars[0][0] < time_min_max[0]:
                min_time_index = np.where(
                    (get_data_vars[0] <= time_min_max[0]))[0][-1]
            else:
                min_time_index = 0
            if time_min_max[1] < get_data_vars[0][-1]:
                max_time_index = np.where(
                    time_min_max[1] <= get_data_vars[0])[0][0]
            else:
                max_time_index = -1
            ylim_min = np.nanmin(
                get_data_vars[1][min_time_index:max_time_index])
            ylim_max = np.nanmax(
                get_data_vars[1][min_time_index:max_time_index])
            ylim(t_plot_name, ylim_min, ylim_max)

    if ('64' in datatype) or ('256' in datatype) or (datatype == 'pot8Hz'):
        for elem in component:
            t_plot_name = prefix+elem
            get_data_vars = get_data(t_plot_name)
            dl_in = get_data(t_plot_name, metadata=True)
            time1 = get_data_vars[0]
            data = np.where(get_data_vars[1] <= -
                            1e+30, np.nan, get_data_vars[1])
            dt = get_data_vars[2]
            ndt = dt.size
            ndata = data.size
            time_new = (np.tile(time1, (ndt, 1)).T + dt * 1e-3).reshape(ndata)
            data_new = data.reshape(ndata)
            store_data(t_plot_name, data={
                       'x': time_new, 'y': data_new}, attr_dict=dl_in)
            options(t_plot_name, 'ytitle', '\n'.join(t_plot_name.split('_')))
            # ylim settings because pytplot.timespan() doesn't affect in ylim.
            # May be it will be no need in future.
            if time_new[0] < time_min_max[0]:
                min_time_index = np.where((time_new <= time_min_max[0]))[0][-1]
            else:
                min_time_index = 0
            if time_min_max[1] < time_new[-1]:
                max_time_index = np.where(time_min_max[1] <= time_new)[0][0]
            else:
                max_time_index = -1
            ylim_min = np.nanmin(data_new[min_time_index:max_time_index])
            ylim_max = np.nanmax(data_new[min_time_index:max_time_index])
            ylim(t_plot_name, ylim_min, ylim_max)

    if datatype == 'pot':
        for elem in component:
            t_plot_name = prefix+elem
            options(t_plot_name, 'ytitle', elem+' potential')

    if datatype == 'spec':
        options(tnames(prefix + '*spectra*'), 'spec', 1)
        options(tnames(prefix + '*spectra*'), 'colormap', 'jet')
        options(tnames(prefix + '*spectra*'), 'zlog', 1)
        ylim(prefix + 'spectra', 0, 100)
        zlim(prefix + 'spectra', 1e-6, 1e-2)
        options(tnames(prefix + '*spectra*'), 'ysubtitle', '[Hz]')
        options(tnames(prefix + '*spectra*'), 'ztitle', '[mV^2/m^2/Hz]')
        for t_plot_name in (tnames(prefix + '*spectra*')):
            options(t_plot_name, 'ytitle', '\n'.join(t_plot_name.split('_')))

    return loaded_data




import numpy as np
from copy import deepcopy

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

def erg_pgs_make_e_spec(data_in):
    """
    Builds energy spectrogram from the particle data structure
    
    Input:
        data_in: dict
            Particle data structure

    Returns:
        Tuple containing: (energy values for the y-axis, spectrogram values)

    """

    data = deepcopy(data_in)

    data_array = deepcopy(data['data'])
    # zero inactive bins to ensure areas with no data are represented as NaN
    data_array = np.where(data['bins'] == 0,0,data['data'])
    ave = data_array.sum(axis=1) / data['bins'].sum(axis=1)

    y = data['energy'][:, 0]

    return (y, ave)
